 defstar {
 	name {MEM_test}
	domain { DE }
	author { Dorgham Sisalem}
        copyright { SEE $PTOLEMY_SIMULATIONS/TCP/copyright }

	desc {
        This star is a TCP source that is based on the 4.3-Tahoe BSD TCP
        implementation and fulfills the following requirements:
        sliding window mechanism: it only sends if the receiver has enough
                 space. Otherwise it waits till the receiverhas freed some
                 buffer and reopened the transmission window.
        timeout: if an acknowledgment for a specific packet has not arrived
                 after some time, it is resent.
        roundtrip time estimation: using the Karns algorithm
        slow start: for the beginning of the transmission and after a timeout
        congestion avoidance: a slow increase of the window size of 1/cwnd
        fast retransmission: after receiving 3 duplicate acks the source 
                             retransmitts the lost packet. 
	}

	input { name { input }
                type { message}
	        desc { new data packets}
        }

	output { name { output }
                 type { message }
	         desc { send data packets}
	}

	defstate {
		name { Win_Size}
		type { int  }
		default { 10 }
		desc {tart value for the window. }
	}
	defstate {
		name { MSS}
		type { int  }
		default { 1 }
		desc {Maximum Segment Size }
	}
	defstate {
		name { init_time}
		type { float }
		default { 1 }
		desc {initial value for RTT}
	}
	defstate {
		name { Rate}
		type { float }
		default { 1 }
		desc {the sending rate of the source}
	}

	hinclude { "TCPPacket.h" ,<math.h>}

	defstate {
		name { fileName }
		type { string }
		default { "<stdout>" }
		desc { Filename for output }
	}
	hinclude { "pt_fstream.h" }
	protected {
		pt_ofstream *p_out;
	}
        //code{ #define MEMORYLOG }

	protected { long count;
		    long Last_Sent,counter;
		    long Limit,state,Retransmission;
		    long Ack;
		    const long STOP=0;
		    const long GO=1;
	            Envelope* buffer;
		    long Last_Ack,Last_Win,Last_Limit;
		    float cwnd,ssthresh,Win;
		    long Slow_Start,Congestion_Avoidance,Time_Out;
		    long Retransmitted,Timer,First_Time;
		    float RTO,A,D,Start_Time;
		    Envelope Envp;
		    long Last_Seq,buffered;
		    long SEQ,Last_Retransmit,send_state;
		    float *timer_buf;
		  }

//*****************************************************************************/
// initialization: cwnd=MSS
//                 start with slow start
//                 RTO is set by the user
//*****************************************************************************/

	setup { delayType                                = FALSE;
	        cwnd                                     = MSS  ;
		Limit                                    = long(cwnd-1);
		state                                    = GO;
		counter                                  = 0;
	        Retransmission                           = 0;
		Ack=Last_Ack=Last_Win                    = -1;
		Last_Limit                               = 0;
		Slow_Start                               = 1;
		Congestion_Avoidance=Time_Out            = 0;
		First_Time                               = 1;
		Timer=Retransmitted                      = -2;
		RTO                                      = init_time;
		Last_Seq                                 = -1;
		send_state                               = 1;
		Win_Size                                 = Win_Size/MSS;
              }

//*****************************************************************************/
// allocate buffer space for the local window
//*****************************************************************************/

	begin  {          //initialize the local window
                if(buffer)
	        {
	          //Error::abortRun(*this,"receiver Buffer overflow","");
                 		  
	                delete [Win_Size] buffer;
                        buffer=NULL;
	         // Error::abortRun(*this,"receiver ","");
		  

		 }
		  //delete buffer;
                
	         buffer=new Envelope[Win_Size];
                //if(p_out) delete p_out;
		LOG_NEW; p_out = new pt_ofstream(fileName);
		}

//****************************************************************************/
// set scheduling priorities
//****************************************************************************/

	constructor { 
                   buffer=NULL;
              }
        wrapup {
	         //if(p_out) LOG_DEL; delete p_out;
	
		if(buffer) 
		  delete [Win_Size] buffer;
		  buffer=NULL;
		}
	destructor { //if(p_out) delete p_out;
		//if(buffer) 
		//  delete [] buffer;
	 }

	go {
                pt_ofstream& OUTPUT = *p_out;

                completionTime= arrivalTime;
//****************************************************************************
//                      SEND NEW DATA
//****************************************************************************

		if(input.dataNew)
		{
	          //read new data and save it in buffer
		  input.get().getMessage(buffer[counter%Win_Size]);
	          const TCPPacket* packet=(const TCPPacket *)
	                      buffer[counter%Win_Size].myData();
		  long num=packet->readNum();
		  long seq=packet->readSeq();
		  output.put(completionTime)<<buffer[counter%Win_Size];

//OUTPUT<<" \n SEND "<<seq<<" at "<<completionTime<<"\n";
		  counter++;

		}

	} // end go
} // end defstar
