 defstar {
 	name {RED}
	domain { DE }
	author { Dorgham Sisalem}
        copyright { SEE $PTOLEMY_SIMULATIONS/copyright }

	desc {This star simulates the random early packet drop scheme. The 
avarage queue length is calculated as an exponential weigthed moving
avaerage. Is the avergae between a Min and a Max threshold then packets
are dropped with a probability b. Is the averge larger than the maximum 
threshold then all arriving packets must be dropped.
	}

	input { name {size }
                type {int }
	}

	input { name {in }
                type {message }
	}

	output { name {out }
                 type {message }
        }
	output { name {overflow }
                 type {message }
        }

	defstate {
		name {Wg }
		type {float }
		default {0.002 }
		desc { queue weight}
	}

	defstate {
		name {Min }
		type {float }
		default {10 }
		desc { min threshold}
	}

	defstate {
		name {Max }
		type {float }
		default {100 }
		desc { max threshold}
	}

	defstate {
		name {MaxP }
		type {float }
		default {0.02 }
		desc { max threshold}
	}
	defstate {
		name {alpha }
		type {float }
		default {0.8 }
		desc { reduction factor}
	}


	hinclude { "TCPPacket.h" ,<math.h>,<Uniform.h>}
	ccinclude{ <ACG.h> }
	protected{ Uniform *random;}
	code{ extern ACG* gen;}

	defstate {
		name { fileName }
		type { string }
		default { "<stdout>" }
		desc { Filename for output }
	}
	hinclude { "pt_fstream.h" }
	protected {
		pt_ofstream *p_out;
	}

	protected { 	float av; 
			int count;
			float C1,C2,Pb;
			double R;
			int Size;
			float min,max;
		  }

//*****************************************************************************/
// initialization: 
//*****************************************************************************/

	setup { av=0;
	        count=-1;
		min=max=0;
		C1=MaxP/(Max-Min);
		C2=(MaxP*Min)/(Max-Min);
              }

//*****************************************************************************/
//*****************************************************************************/

	begin  {  
                LOG_DEL; delete p_out;
		LOG_NEW; p_out = new pt_ofstream(fileName);
		if(random) delete random;
		random=new Uniform(0,1,gen);
		}

//****************************************************************************/
// set scheduling priorities
//****************************************************************************/

	constructor { 
                      p_out = 0;
		      random = NULL;
		      
               	    }
        wrapup {
	        LOG_DEL; delete p_out;
		p_out = 0;
		}
	destructor { LOG_DEL; delete p_out;}

	go {
                pt_ofstream& OUTPUT = *p_out;
                completionTime= arrivalTime;
//**********************************************************
//  Update the queue length average 
//**********************************************************
		if(size.dataNew)
		{
		   	Size=size.get();
		  	float temp=(float)Size-av; 
		   av=av+(float)((float)Wg*(float)temp);
		}
//*********************************************************
//  receive new data
//*********************************************************
		if(in.dataNew)
		{
                        Envelope Envp;
                        in.get().getMessage(Envp);
               		TCPPacket * packetPtr=(TCPPacket *)
                    			Envp.myData();
		if((Min<=av)&&(Max>av))// average between Min and Max
			{
// OUTPUT<<" AVERAGE "<<av<<" Size "<<Size<<" Min "<<Min<<" R/Pb "<<R/Pb<<" cou//nt"<<count<<" at "<<completionTime<<"\n";
                                count++;
				if(count==0)
					R=(*random)();
			        if(!packetPtr->readECN_ON())
			           Pb=(float)alpha*((float)(C1*av)-C2);
				else
				   Pb=(C1*av)-C2;
	                        if((count>0)&&(count>=(R/Pb)))
				{
//OUTPUT<<" Mark DEST "<<packetPtr->readDestination()<<" VC "<<packetPtr->readV//C()<<" SEQ "<<packetPtr->readSeq()<<"\n";
					count=0;
	                                overflow.put(completionTime)<<Envp;
					if(!packetPtr->readECN_ON())
					   return;
					else
                                        { 
                                          TCPPacket* last_Packet = new TCPPacket( count ); //create the packet
					  last_Packet->setDestination(packetPtr->readDestination());
					  last_Packet->setSource(packetPtr->readSource());
					  last_Packet->setVC(packetPtr->readVC());
					  last_Packet->setSize(packetPtr->readSize());
					  last_Packet->setWin(packetPtr->readWin());
					  last_Packet->setNum(packetPtr->readNum()); 
					  last_Packet->setSeq(packetPtr->readSeq());
					  last_Packet->setAck(packetPtr->readAck());
					  last_Packet->setTimestamp(packetPtr->readTimestamp());
					  last_Packet->setECN(1);
					  Envelope outEnvp( *last_Packet);
					  out.put(completionTime)<<outEnvp;
					  return;
					 }
				           
				}
			        Envelope envp(*packetPtr);
	                        out.put(completionTime)<<envp;
			}
			else
			{
				 count=-1;
				if(Max<=av)
				{
//OUTPUT<<" AVERAGE @ "<<" MAX "<<Max<<av<<" at "<<completionTime<<"\n";
			                count=-1;		
	                                overflow.put(completionTime)<<Envp;
					if(!packetPtr->readECN_ON())
					   return;
					else
                                        { 
// OUTPUT<<" AVERAGE "<<av<<" Size "<<Size<<" Max "<<max<<" R/Pb "<<R/Pb<<" cou//nt"<<count<<" at "<<completionTime<<"\n";
//OUTPUT<<" Severe DEST "<<packetPtr->readDestination()<<" VC "<<packetPtr->rea//dVC()<<" SEQ "<<packetPtr->readSeq()<<"\n";
                                          TCPPacket* last_Packet = new TCPPacket( count ); //create the packet
					  last_Packet->setDestination(packetPtr->readDestination());
					  last_Packet->setSource(packetPtr->readSource());
					  last_Packet->setVC(packetPtr->readVC());
					  last_Packet->setSize(packetPtr->readSize());
					  last_Packet->setWin(packetPtr->readWin());
					  last_Packet->setSeq(packetPtr->readSeq());
					  last_Packet->setAck(packetPtr->readAck());
					  last_Packet->setNum(packetPtr->readNum());
					  last_Packet->setTimestamp(packetPtr->readTimestamp());
					  last_Packet->setECN(1);
					  Envelope outEnvp( *last_Packet);
					  out.put(completionTime)<<outEnvp;
					  return;
					 }
                                       
				}
			        Envelope envp(*packetPtr);
	                        out.put(completionTime)<<envp;
			}

		}


	   } // end go
} // end defstar
