defstar {
 	name { ReceiverWin}
	domain { DE }
	author { Dorgham Sisalem }
        copyright { SEE $PTOLEMY_SIMULATIONS/TCP/copyright }

        desc   { This star simulates the buffer of the application.
                 It is assumed that the packets always have a constant size
	         -MSS-. The buffer only receives the sequence number from the
	         TCP receiver. When the application is ready for a new packet 
                 the size of the window is changed bu MSS
	         If there is any interest in variable packet sizes
	         then the receiver must be modified to send real packets.}

	input { name { demand}
                type { int }
	        desc { the application has finished consuming the last packet}
	}
	input { name { inData }
                type { anytype}
	        desc { new data}
	}

	output {
		name {outData}
		type {=inData}
	        desc { send data to the application}
	}
	output {
		name {size}
		type {int}
		desc { Size of Queue. Triggered by demand inputs. }
	}
	output {
		name {overflow}
		type {=inData}
		desc {
	        Arrival data that cannot be queued due to capacity limit.
		}
	}

	protected {
		int infinite;
	}
	defstate {
		name {capacity}
		type {int}
		default {"-1"}
		desc { Maximum size of the queue. If <0, capacity is infinite. }
	}
	defstate {
		name {MSS}
		type {int}
		desc { Maximum segment size}
	}

	constructor {
		// Indicate to the scheduler that size outputs are triggered
		// demand inputs.
		// simultaneous inData events should be available in the
		// same firing with demand events
         	   demand.before(inData);
	}
	setup {
		infinite = (int(capacity) < 0);
		control=1;
		Q_Size=0;
		zapQueue();
	}

	ccinclude {
		"DataStruct.h","TCPPacket.h"
	}
	protected {
		Queue queue;
	}
	method {
		name { enqueue }
		access { protected }
		arglist { "()" }
		type { void }
		code {
		    if (infinite || Qsize() < int(capacity)) {
			queue.put(inData.get().clone());
		    } else {
			// Data is lost
			overflow.put(completionTime) = inData.get();
		    }
		}
	}
	method {
		name { dequeue }
		access { protected }
		arglist { "()" }
		type { Pointer }
		code {
		    return queue.get();
		}
	}
	method {
		name { handleOverflow }
		access { protected }
		arglist { "()" }
		type { void }
		code {
		    // Nothing to do
		}
	}
	method {
		name { Qsize }
		access { protected }
		arglist { "()" }
		type { int }
		code {
		    return queue.length();
		}
	}
	method {
		name { zapQueue }
		code {
			while (Qsize() > 0) {
				Particle* pp = (Particle*) queue.get();
				pp->die();
			}
			queue.initialize();
		      }
               }
	protected{
	  int control;
	  int Q_Size;
	}
	go {
	    completionTime = arrivalTime;

//****************************************************************************
// The application finished consuming a packet an can take another one
//****************************************************************************

	    if (demand.dataNew)
	    {
	      demand.dataNew= FALSE;
       	      if(Qsize())
	      {
		dequeue();
		Q_Size-=MSS;  //update the buffer size
		outData.put(completionTime)<<Q_Size;
		control=0; // application is busy
	      }
	      else
	        control=1; //application is free
	      size.put(completionTime)<<Q_Size; // update the window size
	    }

//*************************************************************************
// If there is new inData, store it if the application is busy
//*************************************************************************

	    if (inData.dataNew)
	    {
	      inData.dataNew=FALSE;
	      if(!Qsize()&&control)
	      {
		outData.put(completionTime) =inData%0;
	        control=0;
	      }
	      else
	      {
		enqueue();
	        Q_Size+=MSS;

	      }
	      size.put(completionTime)<<Q_Size;
	    }

	}

        destructor {
		zapQueue();
	  }
}
