defstar {
 	name { SlidingWin }
	domain { DE }
	author { Dorgham Sisalem }
        copyright { SEE $PTOLEMY_SIMULATIONS/TCP/copyright }

        derivedfrom{RepeatStar}
        desc{
        This star receives the input Packets and sends them to the TCP source
        if this is possible -i.e. the transmission window is not full yet.
        Otherwise it buffers the Packets and waits till the source sends
        a notification, that the window has moved and it is possible to send
        data. In this case it sends the allowed number of bytes.
        }

	input {
                name { refire}
	        type { float }
	        desc {send another packet
	                }
        }
	input {
                name { demand}
	        type { int }
	        desc { number of bytes that can be sent before the transmission
	               window closes }
        }
	input {
                name { inData }
	        type { anytype}
	        desc { new data}
	}

	output {
		name {restart}
		type {float}
	        desc { send another packet}
	}
	output {
		name {outData}
		type {=inData}
	        desc { send data to the TCP source }
	}
	output {
		name {size}
		type {int}
		desc { Size of Queue. }
	}
	output {
		name {overflow}
		type {=inData}
		desc {
	        Arrival data that cannot be queued due to capacity limit.
		}
	}
	protected {
		int infinite;
	}
	defstate {
		name {capacity}
		type {int}
		default {"-1"}
		desc { Maximum size of the queue. If <0, capacity is infinite. }
	}
	defstate {
		name {MSS}
		type {int}
		default {"1"}
		desc {Maximum segment size }
	}
	defstate {
		name {Rate}
		type {float}
		default {"1"}
		desc {the sending rate of the source }
	        }

	constructor {
		// Indicate to the scheduler that size and outputs are 
                // trigged by demand inputs.
		demand.triggers(size);
		inData.triggers(overflow);

         	   demand.before(inData);
		   demand.before(refire);
		   refire.triggers(restart);
		   refire.triggers(outData);
		   inData.triggers(outData);
		   inData.triggers(size);
	}

	ccinclude {
		"DataStruct.h","TCPPacket.h"
	}
	protected {
		Queue queue;
	}
	method {
		name { enqueue }
		access { protected }
		arglist { "()" }
		type { void }
		code {
		    if (infinite || Qsize() < int(capacity)) {
			queue.put(inData.get().clone());
		    } else {
			// Data is lost
			overflow.put(completionTime) = inData.get();
		    }
		}
	}
	method {
		name { dequeue }
		access { protected }
		arglist { "()" }
		type { Pointer }
		code {
		    return queue.get();
		}
	}
	method {
		name { handleOverflow }
		access { protected }
		arglist { "()" }
		type { void }
		code {
		    // Nothing to do
		}
	}
	method {
		name { Qsize }
		access { protected }
		arglist { "()" }
		type { int }
		code {
		    return queue.length();
		}
	}
	method {
		name { zapQueue }
		code {
			while (Qsize() > 0) {
				Particle* pp = (Particle*) queue.get();
				pp->die();
			}
			queue.initialize();
		      }
                }
	protected{
	  int control,Size;
	  float Completion_Time;
	}
	defstate {
		name { fileName }
		type { string }
		default { "<stdout>" }
		desc { Filename for output }
	}
	hinclude { "pt_fstream.h" }
	protected {
		pt_ofstream *p_out;
	}

	setup {
		infinite = (int(capacity) < 0);
		control=MSS;
		Completion_Time=-1;
		Size=MSS;
		zapQueue();
                LOG_DEL; delete p_out;p_out=NULL;
		LOG_NEW; p_out = new pt_ofstream(fileName);
		//DERepeatStar::setup();
	}

        wrapup {
	        LOG_DEL; delete p_out;
		p_out = 0;
		}
	destructor { LOG_DEL; delete p_out;}
        constructor {p_out=NULL;		
	             delayType = TRUE;
                     }
	go {
            pt_ofstream& OUTPUT = *p_out;
	    completionTime = arrivalTime;

//**************************************************************************
//           update the allowed number of byte to be sent
//**************************************************************************

	    if (demand.dataNew)
	    {
	      demand.dataNew= FALSE;
              control=demand%0;    // update the allowed number of 
	  				  // packets to be sent.
              float tmp=completionTime-Completion_Time-Rate+1;
//OUTPUT<<" TMP "<<tmp;
	      tmp=(float)100000000.0*tmp;
//OUTPUT<<" TMP1 "<<tmp;
	      tmp=tmp-(float)100000000.0;
//OUTPUT<<" TMP2 "<<tmp<<" at "<<completionTime<<"\n";
	    if(tmp>10)
	    {
//OUTPUT<<" refire  at "<<completionTime<<"\n";
                  //restart.put(completionTime)<<0;   //schedule the next send
	          refire.dataNew=TRUE;
            }

	    }


//**************************************************************************
// If there is new inData, store it if the window is full
//**************************************************************************

	    if (inData.dataNew)
	    {
	      inData.dataNew=FALSE;
	      if(Qsize())
	      {
	        enqueue();
		return;
	      }
	      if((Size<=control)&&
	            (completionTime>Completion_Time+Rate))
	      {
//OUTPUT<<" NEW RESTART "<<"\n";
	        Envelope Envp;
	        inData.get().getMessage(Envp);
		outData.put(completionTime) <<Envp;
	        refire.dataNew=TRUE;
                //restart.put(completionTime)<<0;   //schedule the next send
	        const TCPPacket* Packet=(const TCPPacket *)Envp.myData();
		Size+=Packet->readSize();
	        Completion_Time=completionTime;
	      }
	      else
		enqueue();
	      size.put(completionTime)<<Qsize();
	    }
//************************************************************************
//   send a new packet
//************************************************************************

	    if(refire.dataNew)
	    {
	      refire.dataNew=FALSE;
     	      if((Size<=control)&&Qsize())
	      {
		Particle* pp =(Particle *)  dequeue();
		Envelope Envp;
		pp->getMessage(Envp);
		const TCPPacket* Packet=(const TCPPacket *)Envp.myData();
		Size+=Packet->readSize();
//OUTPUT<<" OUT Size "<<control<<" Rate "<<Rate<<" at "<<completionTime<<" \n";
		outData.put(completionTime)<<Envp;
                restart.put(completionTime)<<0;//schedule the next send
		pp->die();
	        Completion_Time=completionTime;
	      }
	    }

	}

        destructor {
		zapQueue();
	  }
}
