static const char file_id[] = "DETCP_Rec.pl";
// .cc file generated from DETCP_Rec.pl by ptlang

#ifdef __GNUG__
#pragma implementation
#endif

#include "DETCP_Rec.h"

const char *star_nm_DETCP_Rec = "DETCP_Rec";

const char* DETCP_Rec :: className() const {return star_nm_DETCP_Rec;}

ISA_FUNC(DETCP_Rec,DEStar);

Block* DETCP_Rec :: makeNew() const { LOG_NEW; return new DETCP_Rec;}

DETCP_Rec::DETCP_Rec ()
{
	setDescriptor("This star simulates a TCP receiver that fulfills the following\n        requirements:\n        Sliding window: Using  acknowledgment packets, the receiver informs\n                the source of the allowed number of bytes that can be sent\n	        before the sender buffer gets full.\n        Update packets: After sending an acknowledgment with the window size \n                smaller than the maximum segment size (MSS) an update packet \n                is sent when there is new buffer space available.\n        silly window: If the window size gets too small -smaller than MSS-\n	        an update packet is only sent after freeing enough buffer space\n	        - here, it was set to half the original window size.\n        Duplicate acks: If a packet is lost the receiver sends for each new \n                received packet an Ack for the last correctly received packet.\n                All out of order packets get buffered until the missing packet is\n	        received. Only then can these packets get also acknowledged.\n        Delayed acks: Ack packets are sent after an ack timer goes off or every\n                other packet.\n\n        In this model it is assumed that all packets have the same length\n        -MSS-. Based on this the receiver only informs the application when a\n        packet is received. If there is any interest in using variable packet\n        sizes, the actual received packet must be sent to the application.");
	addPort(input.setPort("input",this,MESSAGE));
	addPort(Ack_Fin.setPort("Ack_Fin",this,INT));
	addPort(Size.setPort("Size",this,INT));
	addPort(control.setPort("control",this,MESSAGE));
	addPort(application.setPort("application",this,INT));
	addPort(Ack_Start.setPort("Ack_Start",this,INT));
	addState(Window_Size.setState("Window_Size",this,"10","maximum transmission window."));
	addState(fileName.setState("fileName",this,"","fileName"));
	addState(MSS.setState("MSS",this,"1","Maximum Segment Size"));
	addState(Ack_Time.setState("Ack_Time",this,"1","Start value for the acknowledgment timer."));
	addState(debug.setState("debug",this,"0","debug"));

# line 153 "DETCP_Rec.pl"
Size.before(Ack_Fin);
  	  Size.before(input);
	  Size.triggers(control);
	  Ack_Fin.before(input);
	  Ack_Fin.triggers(control);
          p_out = NULL;
	  buffer=NULL;
          if(buffer) {LOG_DEL; delete [Window_Size] buffer; buffer=NULL;}
          if(p_out){LOG_DEL; delete p_out; p_out=NULL;}
}

void DETCP_Rec::wrapup() {
# line 146 "DETCP_Rec.pl"
if(buffer) {LOG_DEL; delete [Window_Size] buffer; buffer=NULL;}
	       if(p_out){LOG_DEL; delete p_out;
	       p_out=NULL;}
}

void DETCP_Rec::begin() {
# line 136 "DETCP_Rec.pl"
if(buffer) {LOG_DEL; delete [Window_Size] buffer; buffer=NULL;}
	  LOG_NEW;
	  buffer=new int[Window_Size];
	  for(int i=0;i<Window_Size;i++)
                 buffer[i]=0;
          if(p_out){LOG_DEL; delete p_out; p_out=NULL;}
          LOG_NEW; p_out = new pt_ofstream(fileName);
}

DETCP_Rec::~DETCP_Rec() {
# line 163 "DETCP_Rec.pl"
if(p_out){LOG_DEL; delete p_out; p_out=NULL;}
               if(buffer) {LOG_DEL; delete [Window_Size] buffer;buffer=NULL;}
}

void DETCP_Rec::setup() {
# line 122 "DETCP_Rec.pl"
Ack_Flag=-1;
	  Last_Rec=-1;
	  Last_Num=0;
	  buf_size=0;
	  delayType= FALSE;
	  Win_update=0;
	  Win_Size=Window_Size;
	  Window_Size=(int)Window_Size/(int)MSS;
	  Retransmit=0;
	  counter=pointer=0;
	  ECN=0;
}

void DETCP_Rec::go() {
# line 171 "DETCP_Rec.pl"
completionTime= arrivalTime;
          pt_ofstream& OUTPUT = *p_out;

//***************************************************************************
//                   update the window size
//***************************************************************************

	  if(Size.dataNew)
          {
            int actual_win=int(Size.get());
	   // if(!counter)                   // update only if nothing is 
	      Win_Size=(int)(Window_Size*MSS)-actual_win-((float)counter*(float)MSS);//saved in the local buffer

	    if(Win_update)  //handle silly window and send an update packet
	      if(float(actual_win)/float(Window_Size)<=0.5)
	      {

                const TCPPacket* Packet=(const TCPPacket *)
	                      (theEnvp).myData();
		int seq=Packet->readSeq();
		int num=Packet->readNum();
                last_Packet->setDestination(Packet->readDestination());
                last_Packet->setSource(Packet->readSource());
                last_Packet->setVC(Packet->readVC());
	    	last_Packet->setSize(Packet->readSize());
                last_Packet->setWin(Win_Size);
	        last_Packet->setSeq(Last_Rec);
	        last_Packet->setAck(Last_Num);
		last_Packet->setTimestamp(Last_Rec);
		last_Packet->setECN(Packet->readECN());
	        Envelope outEnvp( *last_Packet);
		Win_update=0;
		control.put(completionTime)<<outEnvp;
       	      }
	  }

//***************************************************************************
//                      new data received
//***************************************************************************

	  if(input.dataNew)
	  {
            //read the new packet
            input.get().getMessage(theEnvp);
	    const TCPPacket* PacketPtr = (const TCPPacket*) 
	                               (theEnvp).myData();
	    last_Packet=new TCPPacket(*PacketPtr);
	    seq = PacketPtr->readSeq();
	    num=PacketPtr->readNum();
	    size=PacketPtr->readSize();
	    VC=PacketPtr->readVC();
	    ECN=PacketPtr->readECN();
if(debug)
OUTPUT<<" SEQ "<<seq<<" NUM "<<num<<" ECN "<<ECN<<" at "<<completionTime<<"\n";            if(Last_Rec>=seq)
	    {
                last_Packet->setNum(num);
                last_Packet->setSeq(seq);
                last_Packet->setVC(VC);
                last_Packet->setWin(Win_Size);
		last_Packet->setECN(PacketPtr->readECN());
                Envelope outEnvp( *last_Packet);
                control.put(completionTime)<<outEnvp;
	 	return;
	
	    }
		
            int STOP=0;
	    if(Last_Rec+1<seq)             //missing packets ??
	    {
	      for(int i=0;(i<counter)&&!STOP;i++)
	      {
	      	if(buffer[i%(int)Window_Size]==(int)seq) 
	           return;              
		if(i+1<counter)
		{
		    if((buffer[i%(int)Window_Size]<seq)&&
			(buffer[(i+1)%(int)Window_Size]>seq))
		    {
			for(int tmp=counter-1;tmp>i;tmp--)
	        	  buffer[(int)(tmp)%(int)Window_Size]=//reorder  packets
				buffer[((int)(tmp)+1)%(int)Window_Size]; 
			STOP=1;
                        buffer[(int)(i+1)%(int)Window_Size]=seq;
		    }
		}
		else
		{
                        buffer[(int)(counter)%(int)Window_Size]=seq;
		}
	      }
	      if(!counter)
	         buffer[(int)(counter)%(int)Window_Size]=seq;


	      if(counter+1>=Window_Size)
	      {
		Error::abortRun(*this,"receiver Buffer overflow","");
		return;
	      }
		counter++;
	        Win_Size-=size;

	        //send a duplicate ack
	        Retransmit++;   //count the number of out of sequence
			      //packets
		last_Packet->setAck(Last_Num);
		last_Packet->setSeq(Last_Rec);
		last_Packet->setVC(VC);
		last_Packet->setWin(Win_Size);
		last_Packet->setTimestamp(Last_Rec);
		last_Packet->setECN(ECN);
		Envelope outEnvp( *last_Packet);
		control.put(completionTime)<<outEnvp;
	        return;
	    }
	    else  // a correct packet was received
	    {
	      Win_Size-=size;
	      if(Retransmit)
	      {
                //send for each out of sequence packet a notification to the
	        //application
	        application.put(completionTime)<<seq;
		buf_size=0;
		last_TS=Last_Rec+1;

//***************************************************************************/
// Control if the buffered packets are correctly received 
//***************************************************************************/

	        while((buffer[(int)pointer%(int)Window_Size]==seq+(++pointer))&&                           (counter!=0))
	        {

	            buf_size+=MSS;
	            application.put(completionTime)<<
                           buffer[(int)(pointer-1)%(int)Window_Size];
		    counter--;
                    Last_Rec=buffer[(int)(pointer-1)%(int)Window_Size];
		}
		pointer--;
		if(!counter) // no packet loss was detected
		  Retransmit=0;

		if(!pointer) // the first packet in the buffer was already 
		  Last_Rec=seq; // out of order, so set the received packet
	                        // as the last correctly received packet

		Ack_Flag=Last_Rec;
		Last_Num=num+buf_size;
		last_Packet->setSeq(Last_Rec);
		last_Packet->setAck(Last_Num);
		last_Packet->setWin(Win_Size);
		last_Packet->setECN(ECN);
		
		int i=0;
		while((buffer[(int)pointer%Window_Size]!=0)&&(pointer!=0)) 
                                                       //reorder the packets
		{                             // that were left in the buffer
		  buffer[(int)(i++)%Window_Size]=
                           buffer[(int)pointer%Window_Size]; 
		  buffer[(int)(pointer++)%Window_Size]=0;
		}
		pointer=0;

	    }
            else
	    {
	      if(seq<=Last_Rec) //a packet has been received that has already
	         Ack_Flag=-2;  // been acknowledged

	      else
	      {
	           application.put(completionTime)<<seq;
		   Last_Rec=seq;
		   Last_Num=num;
		   Retransmit=counter=buf_size=pointer=0;
	      }
	    }

	      if(Ack_Flag!=-1) // a packet has already been received
	      {
		last_TS=seq;
		Last_Win=Win_Size;
	        if(Win_Size<=MSS) Win_update=1;   // silly window ?
	        if(Win_update)
	        {
	            last_Packet->setWin(0);
		    Last_Win=0;
		}
	        else
	     	    last_Packet->setWin(Win_Size);
	     	last_Packet->setTimestamp(last_TS);
		last_Packet->setVC(VC);
		last_Packet->setECN(ECN);
		Envelope outEnvp( *last_Packet);
		control.put(completionTime)<<outEnvp; // send ack
		Ack_Flag=-1;
	      }
	      else  //start ack timer
	      {
	        Ack_Flag=seq;
		last_TS=seq;
		Ack_Start.put(completionTime)<<int(seq);
	      }
	    }
	  }

//*************************************************************************
//                            ack timer went off
//************************************************************************

	  if(Ack_Fin.dataNew)
	  {
	     Ack_Fin.dataNew=FALSE;

	     //if the timer went off for a packet that has not been
	     //acknowledged yet, send an ack

	     if((Ack_Flag==(int)(Ack_Fin%0))&&!Retransmit)
	     {
	        const TCPPacket* Packet=(const TCPPacket *)
	                     (theEnvp).myData();  
	        int seq=Packet->readSeq();
		int num=Packet->readNum();
		Last_Win=Win_Size;
	        if(Win_Size<=MSS) Win_update=1;   // silly window ?
	        if(Win_update)
	        {
	            last_Packet->setWin(0);
		    Last_Win=0;
		}
	        else
	     	    last_Packet->setWin(Win_Size);
                last_Packet->setDestination(Packet->readDestination());
                last_Packet->setSource(Packet->readSource());
                last_Packet->setVC(Packet->readVC());
                last_Packet->setSize(Packet->readSize());
		last_Packet->setVC(VC);
	        last_Packet->setSeq(seq);
	        last_Packet->setAck(num);
	     	last_Packet->setTimestamp(seq);
		last_Packet->setECN(ECN);
	        Envelope outEnvp( *last_Packet);
		control.put(completionTime)<<outEnvp;  // send ack
	        Ack_Flag=-1;
	     }
          }
}

// prototype instance for known block list
static DETCP_Rec proto;
static RegisterBlock registerBlock(proto,"TCP_Rec");
