 defstar {
 	name {TCPfoo }
	domain { DE }
	author { Dorgham Sisalem}
        copyright { SEE $PTOLEMY_SIMULATIONS/TCP/copyright }

	desc {
        This star is a TCP source that is based on the 4.3-Tahoe BSD TCP
        implementation and fulfills the following requirements:
        sliding window mechanism: it only sends if the receiver has enough
                 space. Otherwise it waits till the receiverhas freed some
                 buffer and reopened the transmission window.
        timeout: if an acknowledgment for a specific packet has not arrived
                 after some time, it is resent.
        roundtrip time estimation: using the Karns algorithm
        slow start: for the beginning of the transmission and after a timeout
        congestion avoidance: a slow increase of the window size of 1/cwnd
        fast retransmission: after receiving 3 duplicate acks the source 
                             retransmitts the lost packet. 
	}

	input { name { input }
                type { message}
	        desc { new data packets}
        }

	output { name { output }
                 type { message }
	         desc { send data packets}
	}

	defstate {
		name { Win_Size}
		type { int  }
		default { 10 }
		desc {start value for the window. }
	}
	hinclude { "TCPPacket.h" ,<math.h>}

	defstate {
		name { fileName }
		type { string }
		default { "<stdout>" }
		desc { Filename for output }
	}
	hinclude { "pt_fstream.h" }
	protected {
		pt_ofstream *p_out;
	}

	protected { long count;
		    long int Last_Sent,counter;
		    long int Limit,state,Retransmission;
		    long int Ack;
		    const long int STOP=0;
		    const int GO=1;
	            Envelope* buffer;
		    long int Last_Ack,Last_Win,Last_Limit;
		    float cwnd,ssthresh,Win;
		    long int Slow_Start,Congestion_Avoidance,Time_Out;
		    long int Retransmitted,Timer,First_Time;
		    float RTO,A,D,*Start_Time;
		    Envelope Envp;
		    long int Last_Seq,buffered;
		    long int SEQ,Last_Retransmit,send_state;
		    float *timer_buf,last_TS;
		  }

//*****************************************************************************/
// initialization: cwnd=MSS
//                 start with slow start
//                 RTO is set by the user
//*****************************************************************************/

	setup { delayType                                = FALSE;
		Limit                                    = long (cwnd-1);
		state                                    = GO;
		counter                                  = 0;
	        Retransmission                           = 0;
		Ack=Last_Ack=Last_Win                    = -1;
		Last_Limit                               = 0;
		Slow_Start                               = 1;
		Congestion_Avoidance=Time_Out            = 0;
		First_Time                               = 1;
		Timer=Retransmitted                      = -2;
		Last_Seq                                 = -1;
		send_state                               = 1;
              }

//*****************************************************************************/
// allocate buffer space for the local window
//*****************************************************************************/

	begin  {          //initialize the local window
                if(buffer) LOG_DEL; delete buffer;
	        LOG_NEW;
	        buffer=new Envelope[Win_Size];
                if(timer_buf) LOG_DEL; delete timer_buf;
	        LOG_NEW;
	        timer_buf=new float[Win_Size];
                if(Start_Time) LOG_DEL; delete Start_Time;
	        LOG_NEW;
	        Start_Time=new float[Win_Size];
                LOG_DEL; delete p_out;
		LOG_NEW; p_out = new pt_ofstream(fileName);
		}

//****************************************************************************/
// set scheduling priorities
//****************************************************************************/

	constructor { if(buffer) LOG_DEL; delete buffer;
	              if(timer_buf) LOG_DEL; delete timer_buf;
	              if(Start_Time) LOG_DEL; delete Start_Time;
                      p_out = 0;
              }
        wrapup {
	        LOG_DEL; delete p_out;
		p_out = 0;
		}
	destructor { LOG_DEL; delete p_out;}

	go {
                pt_ofstream& OUTPUT = *p_out;
                completionTime= arrivalTime;

//****************************************************************************
//                      SEND NEW DATA
//****************************************************************************

		if(input.dataNew)
		{
	          //read new data and save it in buffer
	          if(buffer[counter%Win_Size].empty()==0)
	          {
	             OUTPUT<<" count "<<counter<<" Win "<<Win_Size<<" % "<<counter%Win_Size;
		  }
		  input.get().getMessage(buffer[counter%Win_Size]);
	          const TCPPacket* packet=(const TCPPacket *)
	                      buffer[counter%Win_Size].myData();
		  long int num=packet->readNum();
		  long int seq=packet->readSeq();
		  output.put(completionTime)<<buffer[counter%Win_Size];

OUTPUT<<" \n SEND "<<seq<<" RTO "<<RTO<<"\n";
OUTPUT<<" RET5  "<<seq<<"  Win "<<Win_Size;
	          Start_Time[seq%Win_Size]=arrivalTime;
OUTPUT<<" RET4   "<<seq<<"  Win "<<Win_Size;
	          timer_buf[seq%Win_Size]=completionTime+RTO;// register the 
		                                             // time a timeout
							     // should occur
		  Last_Sent=num;
		  counter++;
OUTPUT<<" RETc   "<<counter<<"  Win "<<Win_Size;

		}

	} // end go
} // end defstar
