defstar {
 	name { RecieverWin}
	domain { DE }
	author { Dorgham Sisalem }

	input { name { demand} type { int } }
	input { name { inData } type { anytype} }

	output {
		name {outData}
		type {=inData}
	}
	output {
		name {size}
		type {int}
		desc { Size of Queue. Triggered by demand inputs. }
	}
	output {
		name {overflow}
		type {=inData}
		desc {
Arrival data that cannot be queued due to capacity limit.
		}
	}
	protected {
		int infinite;
	}
	defstate {
		name {capacity}
		type {int}
		default {"-1"}
		desc { Maximum size of the queue. If <0, capacity is infinite. }
	}

	constructor {
		// Indicate to the scheduler that size outputs are triggered
		// demand inputs.
		demand.triggers(size);
		inData.triggers(overflow);

		// simultaneous inData events should be available in the
		// same firing with demand events
         	   demand.before(inData);
	           demand.triggers(outData);
		   inData.triggers(outData);
		   inData.triggers(size);
	}
	setup {
		infinite = (int(capacity) < 0);
		control=1;
		zapQueue();
	}

	ccinclude {
		"DataStruct.h","TCPCell.h"
	}
	protected {
		Queue queue;
	}
	method {
		name { enqueue }
		access { protected }
		arglist { "()" }
		type { void }
		code {
		    if (infinite || Qsize() < int(capacity)) {
			queue.put(inData.get().clone());
		    } else {
			// Data is lost
			overflow.put(completionTime) = inData.get();
		    }
		}
	}
	method {
		name { dequeue }
		access { protected }
		arglist { "()" }
		type { Pointer }
		code {
		    return queue.get();
		}
	}
	method {
		name { handleOverflow }
		access { protected }
		arglist { "()" }
		type { void }
		code {
		    // Nothing to do
		}
	}
	method {
		name { Qsize }
		access { protected }
		arglist { "()" }
		type { int }
		code {
		    return queue.length();
		}
	}
	method {
		name { zapQueue }
		code {
			while (Qsize() > 0) {
				Particle* pp = (Particle*) queue.get();
				pp->die();
			}
			queue.initialize();
		      }
                }
	protected{
	  int control;
	}
	go {
	    completionTime = arrivalTime;
	    if (demand.dataNew)
	    {
	      demand.dataNew= FALSE;
       	      if(!Qsize())
	      {
		Particle* pp =(Particle *)  dequeue();
		outData.put(completionTime)=*pp;
		pp->die();
	      }
	      size.put(completionTime)<<Qsize();
	    }

	    // If there is new inData, store it
	    if (inData.dataNew)
	    {
	      inData.dataNew=FALSE;
	      if(!Qsize())
	      {
	        Envelope Envp;
	        inData.get().getMessage(Envp);
		outData.put(completionTime) <<Envp;
	      }
	      else
		enqueue();
	      size.put(completionTime)<<Qsize();
	    }

	}

        destructor {
		zapQueue();
	  }
}
