 defstar {
 	name {TCPSender }
	domain { DE }
	author { Dorgham Sisalem}
        copyright { SEE $PTOLEMY_SIMULATIONS/TCP/copyright }

	desc {
        This star is a TCP source that is based on the 4.3-Tahoe BSD TCP
        implementation and fulfills the following requirements:
        sliding window mechanism: it only sends if the receiver has enough
                 space. Otherwise it waits till the receiverhas freed some
                 buffer and reopened the transmission window.
        timeout: if an acknowledgment for a specific packet has not arrived
                 after some time, it is resent.
        roundtrip time estimation: using the Karns algorithm
        slow start: for the beginning of the transmission and after a timeout
        congestion avoidance: a slow increase of the window size of 1/cwnd
        fast retransmission: after receiving 3 duplicate acks the source 
                             retransmitts the lost packet. 
	}

	input { name { ack}
                type { message}
	        desc { Acknowoledgment packet}
	}
	input { name { input }
                type { message}
	        desc { new data packets}
        }
	input { name { timer}
                type { message}
	}

	output { name { demand}
                 type { int}
	         desc { get demand number of bytes from the WindowBuf}
        }
	output { name { output }
                 type { message }
	         desc { send data packets}
	}
	output { name { timeout}
                 type { int}
	         desc { last calculated round trip time}
	}

	defstate {
		name { Win_Size}
		type { int  }
		default { 10 }
		desc {tart value for the window. }
	}
	defstate {
		name { MSS}
		type { int  }
		default { 1 }
		desc {Maximum Segment Size }
	}
	defstate {
		name { init_time}
		type { float }
		default { 1 }
		desc {initial value for RTT}
	}
	defstate {
		name { Rate}
		type { float }
		default { 1 }
		desc {the sending rate of the source}
	}

	hinclude { "TCPPacket.h" ,<math.h>}

	defstate {
		name { fileName }
		type { string }
		default { "<stdout>" }
		desc { Filename for output }
	}
	hinclude { "pt_fstream.h" }
	protected {
		pt_ofstream *p_out;
	}

	protected { long count;
		    long int Last_Sent,counter;
		    long int Limit,state,Retransmission;
		    long int Ack;
		    const long int STOP=0;
		    const int GO=1;
	            Envelope* buffer;
		    long int Last_Ack,Last_Win,Last_Limit;
		    float cwnd,ssthresh,Win;
		    long int Slow_Start,Congestion_Avoidance,Time_Out;
		    long int Retransmitted,Timer,First_Time;
		    float RTO,A,D,*Start_Time;
		    Envelope Envp;
		    long int Last_Seq,buffered;
		    long int SEQ,Last_Retransmit,send_state;
		    float *timer_buf;
		  }

//*****************************************************************************/
// initialization: cwnd=MSS
//                 start with slow start
//                 RTO is set by the user
//*****************************************************************************/

	setup { delayType                                = FALSE;
	        cwnd                                     = MSS  ;
		Limit                                    = long (cwnd-1);
		state                                    = GO;
		counter                                  = 0;
	        Retransmission                           = 0;
		Ack=Last_Ack=Last_Win                    = -1;
		Last_Limit                               = 0;
		Slow_Start                               = 1;
		Congestion_Avoidance=Time_Out            = 0;
		First_Time                               = 1;
		Timer=Retransmitted                      = -2;
		RTO                                      = init_time;
		Last_Seq                                 = -1;
		send_state                               = 1;
              }

//*****************************************************************************/
// allocate buffer space for the local window
//*****************************************************************************/

	begin  {          //initialize the local window
                if(buffer) LOG_DEL; delete buffer;
	        LOG_NEW;
	        buffer=new Envelope[Win_Size];
                if(timer_buf) LOG_DEL; delete timer_buf;
	        LOG_NEW;
	        timer_buf=new float[Win_Size];
                if(Start_Time) LOG_DEL; delete Start_Time;
	        LOG_NEW;
	        Start_Time=new float[Win_Size];
                LOG_DEL; delete p_out;
		LOG_NEW; p_out = new pt_ofstream(fileName);
		}

//****************************************************************************/
// set scheduling priorities
//****************************************************************************/

	constructor { if(buffer) LOG_DEL; delete buffer;
	              if(timer_buf) LOG_DEL; delete timer_buf;
	              if(Start_Time) LOG_DEL; delete Start_Time;
	              ack.before(input);
		      ack.before(timer);
		      input.before(timer);
                      p_out = 0;
              }
        wrapup {
	        LOG_DEL; delete p_out;
		p_out = 0;
		}
	destructor { LOG_DEL; delete p_out;}

	go {
                pt_ofstream& OUTPUT = *p_out;

                completionTime= arrivalTime;
		if(ack.dataNew)  // an acknowlegment has arrived
 		{
		  ack.get().getMessage(Envp);
		  const TCPPacket* packetPtr=(const TCPPacket *)
		    Envp.myData();
		  Ack=packetPtr->readAck();
		  long int Seq=packetPtr->readSeq();
		  Win=packetPtr->readWin();
OUTPUT<<" seq "<<Seq<<" Ack "<<Ack;
//***************************************************************************/
// ignore Acks for falsely retransmitted packets -i.e. packets that are
// retransmitted after a timeout even though the ack is on the way, or 
// duplicate acks after the 3rd. one was received.
//***************************************************************************/

		  if((Ack<Last_Ack)||((Retransmission>2)&&(Ack==Last_Ack)))
	           {
                      Ack=Last_Ack;
		      return;
		   }

//***************************************************************************/
// If an ack arrives after retransmitting a lost packet, and the ack only
// acknowledges a part of the packets that were already transmitted after 
// the lost packet, then the receiver is indicating that among the packets
// that were buffered in the TCP window at the receiver a loss was detected.
// The source has to retransmit the lost packet directly.
//***************************************************************************/


	          if(Retransmission&&(Ack<Last_Sent)&&(Ack!=Last_Ack))
                  {
OUTPUT<<" Ret "<<Seq;
	                Last_Ack=Ack;
	                timer_buf[(Seq+1)%Win_Size]=completionTime+RTO;
	                Start_Time[(Seq+1)%Win_Size]=completionTime;
		        output.put(completionTime)<<buffer[(Seq+1)%Win_Size];
		        return;
	          }

//***************************************************************************/
// measure round trip time only if there was no timeout for this sequence
// number.
//***************************************************************************/

		  if((Timer!=-2)&&(Ack>Last_Ack))
	          {
	           if(++send_state%2)
	           {
	             float M=arrivalTime-Start_Time[Seq%Win_Size]
		               +packetPtr.readTimestamp();
	             if(First_Time)   // calculate the first RTO
		     {
		       A=M+3;
		       D=A/2;
		       First_Time=0;
		     }
		     else
		     {               // calculate RTO
		       float  Err=M-A;
		       A=A+(0.125*Err);
		       if(Err<0)
		          Err*=-1;
		       D=D+(0.25*(Err-D));
		     }
		     RTO=A+(4*D);

		     if(RTO>64)     //set an upper limit of 64
		        RTO=64;
		     //send latest round trip estimation

		     timeout.put(completionTime)<<RTO;
OUTPUT<<" SEQ "<<Seq<<" Time "<<Start_Time[Seq%Win_Size]<<" RTO "<<RTO<<" at "<<completionTime<<" M "<<M<<" D "<<D<<" A "<<A<<" \n";
                   }
	           
	          }

//***************************************************************************/
// increase cwnd by the size of one packet for each acknowledged packet
//***************************************************************************/
	          if(Slow_Start)
		  {
		     cwnd=2*(Seq-Last_Seq);
		     cwnd*=MSS;
		     if(cwnd<=0) cwnd=MSS;

		  }

//***************************************************************************/
// increase cwnd by the size of one packet for each RTT while in the
// congestion avoidance state.
//***************************************************************************/

		  if(cwnd<Win)
	          {
	              if((cwnd>=ssthresh)&&Congestion_Avoidance)
		      {
		        Slow_Start=0;
			float cwnd_pack=cwnd/(float)MSS;
			cwnd_pack+=1/cwnd_pack;
		        cwnd_pack*=MSS;
			cwnd=cwnd_pack;
			state=1;
		      }
		      Win=cwnd;
		  }

		  Limit=Win+Ack;  //update the sliding window

	OUTPUT<<" win "<<Win;      

//***************************************************************************/
// if 3 duplicet Acks were received then retransmit the lost packet and go into
// slow start.
//***************************************************************************/
		  if((Ack==Last_Ack)&&Last_Win) //duplicate Acks ?
	          {
	              Retransmission++;
	              if(Retransmission<3)
		         return;
	              if(Retransmission==3)
		      {
			Timer=-2;         //stop the RTO calculation
		        state=0;
			Last_Limit=0;
			

			//As a timeout for a packet that was sent after
			//the lost packet is meaningless at this stage
			//clear all time stamps
			long int i;
			for(i=(Seq+1)%Win_Size;i<Win_Size;i++)
		            timer_buf[i]=0;

			//update the time when a timeout shoul occur
	                timer_buf[(Seq+1)%Win_Size]=completionTime+RTO;
	                Start_Time[(Seq+1)%Win_Size]=completionTime;
OUTPUT<<" RET "<<Seq+1;
		        // read the lost packet and send it
		        output.put(completionTime)<<buffer[(Seq+1)%Win_Size];
		        demand.put(completionTime)<<STOP;
		        Last_Seq=counter-1;

			Slow_Start=1;  //go into slow start
 		        Congestion_Avoidance=1; //start the congestion
		                                //avoidance state after
					        //the slow start

			if(cwnd<Win)   //reset the ssthresh value
		        {
		           ssthresh=cwnd/2;
			   ssthresh-=(long int)ssthresh%MSS;
		        }
		        else
		        {
		           ssthresh=Win/2;
			   ssthresh-=(long int)ssthresh%MSS;
		        }
		        if(ssthresh<MSS) ssthresh=MSS+MSS; // set the lower
		                                           // limit of ssthresh
			cwnd=MSS;           // initialize cwnd
		        return;
		       
		      }
	          }
	          else
	          {
                    Last_Ack=Ack;
		    Last_Win=Win;
		    if(Retransmission)
		    {
		      Retransmission=0;
		    }
		   
		    if(Retransmitted>=Seq)
		         Time_Out=0;      //reset the Time_Out state
		  }
		  if(Win<MSS) // avoid silly windows
		  {
		    demand.put(completionTime)<<STOP;
		  }
	
		  if(Limit>=Last_Limit)  //update the sliding window size
		  {
	            Last_Limit=Limit;
		    demand.put(completionTime)<<Limit; // inform the buffer of
		                                       // the possibility to
						       // send
//		    if(Limit>Last_Sent+MSS)
//		      state=GO;

		  }

		
 	      }

//****************************************************************************
//                      SEND NEW DATA
//****************************************************************************

		if(input.dataNew)
		{
	          //read new data and save it in buffer
		  input.get().getMessage(buffer[counter%Win_Size]);
	          const TCPPacket* packet=(const TCPPacket *)
	                      buffer[counter%Win_Size].myData();
		  long int num=packet->readNum();
		  long int seq=packet->readSeq();
		  output.put(completionTime)<<buffer[counter%Win_Size];

		  if(num>=Limit-MSS) //sliding window full?
		  {
		    demand.put(completionTime)<<STOP;
		  }
		  if(Timer==-2) // if no RTT is being measured, start a new
	                        // measurement
	          {
		    Timer=seq;
		  }
OUTPUT<<" \n SEND "<<seq<<" RTO "<<RTO<<" at "<<completionTime<<"\n";
	          Start_Time[seq%Win_Size]=arrivalTime;
	          timer_buf[seq%Win_Size]=completionTime+RTO;// register the 
		                                             // time a timeout
							     // should occur
		  Last_Sent=num;
		  counter++;

		}

//****************************************************************************
//                       timeout for a packet has occured
//****************************************************************************

		if(timer.dataNew)
		{
	          Envelope envp;
		  timer.get().getMessage(envp);
	          const TCPPacket* packet=(const TCPPacket *)
	                      envp.myData();
	          long int seq=packet->readSeq();
		  long int num=packet->readNum();

	          if(num>Last_Ack) // has this packet already been acknowledged
	          {
		    //Ckeck the feasability of the timeout
		    //if the value for this packet was set to 0 or is bigger
		    //than the current time then there is no need for a timeout
	             if((completionTime+Rate<timer_buf[seq%Win_Size])
		                  ||(timer_buf[seq%Win_Size]<Rate))
		        return;
		     Retransmitted=seq; // register the packet that got lost

		     // clear all time stamps to avoid meaningless timeouts
OUTPUT<<" TIMEOUT "<<seq<<" at "<<completionTime;
		     long int i;
		     for(i=seq%Win_Size;i<Win_Size;i++)
		            timer_buf[i]=0;
	             Retransmission=3;
		     Last_Seq=counter-1;// register the packet to be sent
		     Timer=-2;         //stop the RTO calculation
		     state=0;
		     Last_Limit=0;
		     Slow_Start=1;  //go into slow start
		     Congestion_Avoidance=1; //start the congestion
		                                //avoidance state after
					        //the slow start
		     if(cwnd<Win)   //reset the ssthresh value
		     {
		           ssthresh=cwnd/2;
			   ssthresh-=(long int)ssthresh%MSS;
		     }
		     else
		     {
		           ssthresh=Win/2;
			   ssthresh-=(long int)ssthresh%MSS;
		      }
		      if(ssthresh<MSS) ssthresh=MSS+MSS; // set the lower
		                                           // limit of ssthresh
		      cwnd=MSS;                          // initialize cwnd
		      demand.put(completionTime)<<STOP;
		      Time_Out=1;
		      RTO*=2;     // exponential backoff
		      if(RTO>64)
		        RTO=64;
		      timeout.put(completionTime)<<RTO; // update the time out
		                                      // value
							  
		      //calculate the new time stamp for the retransmitted 
		      //packet
	              timer_buf[seq%Win_Size]=completionTime+RTO;
			                    
		      //resend the packet
		      output.put(completionTime)<<envp;
		  }
		}
	} // end go
} // end defstar
