// filename: TCPPacket.h
//
// This file describes the contents of a header of a TCP packet. The TCP 
// Packet idea, as well as much of the following code, was taken from
// C++ PRIMER by Stanley B. Lippman, Addison Weslsey, 2nd ed., 1991.

#ifndef TCPPacket_h
#define TCPPacket_h
#include "Message.h"


#define TCPPACKET_HEADERLENGTH 20         // length of header in bytes
static const int BISDNPacketSize = 8;
 
struct TCPP {
                  long int Seq;
                  long int Ack;
                  long int Win;
                  int Size;
		  int ECN;
		  int ECN_ON;
                  long int Num;
	          double Timestamp;
                  int dummy1;     // not used
                  int dummy2;     // not used
                  int dummy3;     // not used
                };

class TCPPacket : public Message {
public:
//static const char file_id[] = "TCPPacket.h";



	// default constructor
	TCPPacket( const int sequenceNo ) : seqNo( sequenceNo ),
		size( BISDNPacketSize )
	{
		LOG_NEW;
		Packet = new unsigned short;
	}

	// copy constructor
	TCPPacket( const TCPPacket& ba ):size( BISDNPacketSize )
	{
		const int wordWidth = ( size + 8*sizeof(short) - 1 ) /
			8*sizeof(short);
		LOG_NEW;
		Packet = new unsigned short;
		this->numNo = ba.numNo;
		this->ackNo = ba.ackNo;
		this->seqNo = ba.seqNo;
		this->source = ba.source;
		this->destination = ba.destination;
		this->VC = ba.VC;
		this->Time = ba.Time;
		this->LAckNo = ba.LAckNo;
		this->sizeNo = ba.sizeNo;
		this->winNo = ba.winNo;
	}

//		TCPPacket( const int sequenceNo = 0 );
//		TCPPacket( const TCPPacket& ); // copy constructor
	~TCPPacket() { LOG_DEL; delete Packet; }

	StringList print() const {
		StringList out;
		out << "\nTCPPacket: seqNo = ";
		out << (double)readSeq() << "\n";
       		out << "source= "<< readSource()<<"  destination= "
       		     << readDestination() << "VC= "<< readVC()<<'\n';
		return out;
	}
	void setSeq( long int num ) { seqNo = num; }
	void setECN( int num ) { ECN = num; }
	void setECN_ON( int num ) { ECN_ON = num; }
	void setAck( long int num ) { ackNo = num; }
	void setWin( long int num ) { winNo = num; }
	void setSize( long int num ) { sizeNo = num; }
	void setNum( long int num ) { numNo = num; }
	void setLAck( long int num ) { LAckNo = num; }
	void setTimestamp( double num ) { Time = num; }

	long int readSeq() const { return( seqNo ); }
	long int readECN_ON() const { return( ECN_ON ); }
	long int readECN() const { return( ECN ); }
	long int readAck() const { return( ackNo ); }
	long int readWin() const { return( winNo ); }
	long int readSize() const { return( sizeNo ); }
	double readTimestamp() const { return( Time ); }
	long int readNum() const { return( numNo ); }
	long int readLAck() const { return( LAckNo ); }
        int readSource() const { return( source ); }
        int readDestination() const { return( destination ); }
        int readVC() const { return( VC ); }

	void setSource( int num ) { source= num; }
        void setDestination( int num ) { destination= num; }
        void setVC( int num ) { VC= num; }

	const char* dataType() const { return "TCPPacket"; }
	//int isA(const char*) const;
	double asFloat() const { return( double( seqNo ) ); }
	// class identification
	ISA_INLINE(TCPPacket,Message);
	Message* clone() const {
		LOG_NEW; return new TCPPacket( *this );
	}

private:
	unsigned short* Packet;
	long int seqNo;
	int ECN;
	int ECN_ON;
	long int ackNo;
	long int winNo;
	long int sizeNo;
	long int numNo;
	long int LAckNo;
        double Time;
	const int size; // number of bits desired in TCPPacket

        // the following two fields are included to make
        // routing and simulation debugging easier

        int source;     // Number of source node
        int destination;// Number of destination node
        int VC;// Number of VC node
};

#endif
